﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Timers;
using System.Threading;

namespace AT89C2051Programmer
{
    public partial class SerialPortForm : Form
    {
        private Form1 parent = null;
        private System.Timers.Timer timer = null;

        private byte[] rxBuffer = null;
        private int rxInIndex = 0;

        private callbackDelegate callback;

        public SerialPortForm(Form1 parentForm)
        {
            InitializeComponent();
            parent = parentForm;
            Icon = parent.Icon;

            serialPort1.BaudRate = 19200;
            serialPort1.Parity = System.IO.Ports.Parity.None;
            serialPort1.StopBits = System.IO.Ports.StopBits.One;

            timer = new System.Timers.Timer();
            timer.Elapsed += Timer_Elapsed;
            timer.Enabled = false;
            timer.AutoReset = false;
            txLabel.Text = "";
            rxLabel.Text = "";

            rxBuffer = new byte[256];
            rxInIndex = 0;
        }

        public void setPort(string name)
        {
            try
            {
                serialPort1.PortName = name;
            }
            catch (System.ArgumentException)
            {

                serialPort1.PortName = "COM1";
            }
        }

        private void SerialPortForm_Resize(object sender, EventArgs e)
        {
            if (Width > 80)
            {
                txLabel.Width = Width - 40;
                rxLabel.Width = Width - 40;
            }
            else
            {
                txLabel.Width = 40;
                rxLabel.Width = 40;
            }

            if (Height > 137)
            {
                txLabel.Height = (Height - 57) / 2;
                rxLabel.Height = (Height - 57) / 2;
            }
            else
            {
                txLabel.Height = 40;
                rxLabel.Height = 40;
            }
            rxLabel.Top = txLabel.Top + txLabel.Height;
        }

        private void SerialPortForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            timer.Stop();
            timer.Dispose();

            if (serialPort1.IsOpen)
            {
                e.Cancel = true; //cancel the form closing
                Thread CloseDown = new Thread(new ThreadStart(CloseSerialOnExit));
                CloseDown.Start(); //close port in new thread to avoid hang
            }

        }

        private void CloseSerialOnExit()
        {
            try
            {
                serialPort1.Close();
            }

            catch (Exception ex)
            {
                MessageBox.Show(ex.Message); //catch any serial port closing error messages
            }
            try
            {
                this.Invoke(new EventHandler(NowClose)); //now close back in the main thread
            }
            catch (InvalidOperationException)
            {
            }
        }

        private void NowClose(object sender, EventArgs e)
        {
            this.Close(); //now close the form
        }

        public delegate void callbackDelegate(bool success, byte[] buffer);

        public void send(byte[] bytes, int mStimeout, callbackDelegate d)
        {
            byte[] buffer = new byte[2 + bytes.Length * 2];
            int i = 0;
            buffer[i++] = 0x02; // begin packet with STX
            foreach (byte b in bytes)
            {
                if (b >= 0x01 && b <= 0x03) // convert the bytes 01, 02, and 03 to the pairs 01FE, 01FD, and 01FC respectively
                {
                    buffer[i++] = 0x01;
                    buffer[i++] = (byte)(b ^ 0xFF);
                }
                else
                {
                    buffer[i++] = b;
                }
            }
            buffer[i++] = 0x03; // end packet with ETX

            callback = d;
            if (!serialPort1.IsOpen)
            {
                serialPort1.Open();
            }
            timer.Interval = mStimeout;
            timer.Enabled = true;
            rxInIndex = 0;
            serialPort1.Write(buffer, 0, i);
            if (Visible)
            {
                string s = "";
                foreach (byte b in buffer)
                {
                    s += b.ToString("X2") + " ";
                }
                Utils.SetControlPropertyValue(txLabel, "Text", txLabel.Text + s);
            }
        }

        private void Timer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            timer.Stop();
            callback(false, null);
        }

        private void serialPort1_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            rxInIndex += serialPort1.Read(rxBuffer, rxInIndex, serialPort1.BytesToRead);

            // make sure packet begins with an STX by stripping any characters preceding one
            while (rxInIndex > 0 && rxBuffer[0] != 0x02)
            {
                for (int i = 1; i < rxInIndex; i++)
                {
                    rxBuffer[i - 1] = rxBuffer[i];
                }
                rxInIndex--;
            }

            
            if ((rxInIndex > 1) && rxBuffer[rxInIndex - 1] == 0x03) // when ETX received
            {
                timer.Stop();
                // replace any byte pairs with their single byte equivalent
                for (int i = 1; i < rxInIndex; i++)
                {
                    if (rxBuffer[i - 1] == 0x01)
                    {
                        rxBuffer[i - 1] = (byte)(rxBuffer[i] ^ 0xFF);
                        for (int j = i + 1; j < rxInIndex; j++)
                        {
                            rxBuffer[j - 1] = rxBuffer[j];
                        }
                        rxInIndex--;
                    }
                }

                if (Visible)
                {
                    string s = "";
                    for (int i = 0; i < rxInIndex; i++)
                    {
                        s += rxBuffer[i].ToString("X2") + " ";
                    }
                    Utils.SetControlPropertyValue(rxLabel, "Text", rxLabel.Text + s);
                }

                // strip STX and ETX from message
                byte[] buffer = new byte[rxInIndex - 2];
                for (int i = 0; i < rxInIndex - 2; i++)
                {
                    buffer[i] = rxBuffer[i + 1];
                }
                callback(true, buffer);
            }
        }
    }
}
